import six
import numpy as np


class BadBinnedDataframeConfig(Exception):
    pass


def create_binning_list(name, bin_list, make_bins=None):
    if not isinstance(bin_list, list):
        raise BadBinnedDataframeConfig("binning section for stage '{}' not a list".format(name))
    ins = []
    outs = []
    binnings = []
    indices = []
    for i, one_bin_dimension in enumerate(bin_list):
        if not isinstance(one_bin_dimension, dict):
            raise BadBinnedDataframeConfig("binning item no. {} is not a dictionary".format(i))
        cleaned_dimension_dict = {"_" + k: v for k, v in one_bin_dimension.items()}
        _in, _out, _bins, _index = create_one_dimension(name,
                                                        make_bins=make_bins,
                                                        **cleaned_dimension_dict)
        ins.append(_in)
        outs.append(_out)
        indices.append(_index)
        binnings.append(_bins)
    return ins, outs, binnings


def create_one_dimension(stage_name, _in, _out=None, _bins=None, _index=None, make_bins=None):
    if not make_bins:
        make_bins = bin_one_dimension
    if not isinstance(_in, six.string_types):
        msg = "{}: binning dictionary contains non-string value for 'in'"
        raise BadBinnedDataframeConfig(msg.format(stage_name))
    if _out is None:
        _out = _in
    elif not isinstance(_out, six.string_types):
        msg = "{}: binning dictionary contains non-string value for 'out'"
        raise BadBinnedDataframeConfig(msg.format(stage_name))
    if _index and not isinstance(_index, six.string_types):
        msg = "{}: binning dictionary contains non-string and non-integer value for 'index'"
        raise BadBinnedDataframeConfig(msg.format(stage_name))

    if _bins is None:
        bin_obj = None
    elif isinstance(_bins, dict):
        bin_obj = make_bins(**_bins)
        if bin_obj is None:
            msg = "{}: No way to infer binning edges for in={}"
            raise BadBinnedDataframeConfig(msg.format(stage_name, _in))
    else:
        msg = "{}: bins is neither None nor a dictionary for in={}"
        raise BadBinnedDataframeConfig(msg.format(stage_name, _in))

    return (str(_in), str(_out), bin_obj, _index)


def bin_one_dimension(low=None, high=None, nbins=None, edges=None,
                      overflow=True, underflow=True):
    # - bins: {nbins: 6 , low: 1  , high: 5 , overflow: True}
    # - bins: {edges: [0, 200., 900], overflow: True}
    if all([x is not None for x in (nbins, low, high)]):
        bin_obj = np.linspace(low, high, nbins + 1)
    elif edges:
        # array are fixed to float type, to be consistent with the float-type underflow and overflow bins
        bin_obj = np.array(edges, "f")
    else:
        return None
    if underflow:
        bin_obj = np.insert(bin_obj, 0, float("-inf"))
    if overflow:
        bin_obj = np.append(bin_obj, float("inf"))
    return bin_obj


def create_weights(stage_name, weights):
    if weights is None:
        return {}
    if isinstance(weights, list):
        weights = {str(w): w for w in weights}
    elif isinstance(weights, dict):
        pass
    else:
        # else we've got a single, scalar value
        weights = {weights: weights}
    return weights



"""Stages to remove events from subsequent stages

Provides two stages:

  * :class:`CutFlow` -- Prevent subsequent stages from seeing certain events,
  * :class:`SelectPhaseSpace` -- Create a new variable which can be used as a mask

Both stages are configured very similarly, and both stages produce an output
table describing how many events pass each subsequent cut to make it into the
final mask.
"""
from __future__ import absolute_import
import six
import pandas as pd
import os
from copy import deepcopy
from .filters import build_selection


__all__ = ["CutFlow", "SelectPhaseSpace"]


class BadCutflowConfig(Exception):
    pass


class BadSelectionFile(Exception):
    pass


class Collector():
    def __init__(self, filename, keep_unique_id):
        self.filename = filename
        self.keep_unique_id = keep_unique_id

    def collect(self, dataset_readers_list):
        if len(dataset_readers_list) == 0:
            return None

        output = self._prepare_output(dataset_readers_list)
        output.to_csv(self.filename, float_format="%.17g")

    def _prepare_output(self, dataset_readers_list):
        dataset_readers_list = [(d, [r.selection for r in readers])
                                for d, readers in dataset_readers_list
                                if readers]
        if len(dataset_readers_list) == 0:
            return None

        return _merge_data(dataset_readers_list, self.keep_unique_id)


def _merge_data(dataset_readers_list, keep_unique_id=False):
    all_dfs = []
    keys = []
    for dataset, counters in dataset_readers_list:
        output = deepcopy(counters[0])
        for counter in counters[1:]:
            output.merge(counter)
        keys.append(dataset)
        all_dfs.append(output.to_dataframe())

    final_df = pd.concat(all_dfs, keys=keys, names=['dataset'], sort=False)
    if not keep_unique_id:
        final_df.index = final_df.index.droplevel(level="unique_id")

    return final_df


def _load_selection_file(stage_name, selection_file):
    import yaml
    with open(selection_file, "r") as infile:
        cfg = yaml.load(infile)
    if len(cfg) > 1:
        msg = "{}: Selection file has more than one selection"
        raise BadSelectionFile(msg.format(stage_name, selection_file, cfg.keys()))
    return cfg


def _create_weights(stage_name, weights):
    if weights is None:
        return {}
    if isinstance(weights, six.string_types):
        return {weights: weights}

    if isinstance(weights, (tuple, list)):
        weights = {w: w for w in weights}
    if isinstance(weights, dict):
        non_strings = tuple(filter(lambda x: not isinstance(x, six.string_types), weights.values()))
        if non_strings:
            msg = "{}: weight not all string, '{}'"
            raise BadCutflowConfig(msg.format(stage_name, non_strings))
        return weights

    raise BadCutflowConfig("{}: Cannot process weight specification".format(stage_name))


class CutFlow(object):
    """Prevents subsequent stages seeing certain events.

    The two most important  parameters to understand are the ``selection`` and
    ``weights`` parameters.

    Parameters:
      selection (str or dict): The criteria for selecting events, formed by a
        nested set of "cuts".  Each cut must either be a valid :ref:`expressions`
        or a single-length dictionary, with one of ``Any`` or ``All`` as the key,
        and a list of cuts as the value.
      weights (str or list[str], dict[str, str]): How to weight events in the
        output summary table.  Must be either a single variable, a list of
        variables, or a dictionary where the values are variables in the data and
        keys are the column names that these weights should be called in the
        output tables.

    Example:
      Mask events using a single cut based on the ``nJet`` variable being
      greater than 2 and weight events in the summary table by the
      ``EventWeight`` variable::

         cut_flow_1:
             selection:
                 nJet > 2
             weights: EventWeight

      Mask events by requiring both the ``nMuon`` variable being greater than 2
      and the first ``Muon_energy`` value in each event being above 20.  Don't weight
      events in the summary table::

         cut_flow_2:
             selection:
                 All:
                   - nMuon > 2
                   - {reduce: 0, formula: Muon_energy > 20}

      Mask events by requiring the ``nMuon`` variable be greater than 2 and
      either the first ``Muon_energy`` value in each event is above 20 or the
      ``total_energy`` is greater than 100.  The summary table will weight
      events by both the EventWeight variable (called weight_nominal in the
      table) and the SystUp variable (called weight_syst_up in the summary)::

         cut_flow_3:
             selection:
                 All:
                   - nMuon > 2
                   - Any:
                     - {reduce: 0, formula: Muon_energy > 20}
                     - total_energy > 100
             weights: {weight_nominal: EventWeight, weight_syst_up: SystUp}


    Other Parameters:
      name (str):  The name of this stage (handled automatically by fast-flow)
      out_dir (str):  Where to put the summary table (handled automatically by
          fast-flow)
      selection_file (str): Deprecated
      keep_unique_id (bool): If ``True``, the summary table will contain a
          column that gives each cut a unique id.  This is used internally to
          maintain the cut order, and often will not be useful in subsequent
          manipulation of the output table, so by default this is removed.
      counter (bool): Currently unused

    Raises:
      BadCutflowConfig: If neither or both of ``selection`` and
          ``selection_file`` are provided, or if a bad selection or
          weight configuration is given.


    See Also:
      :class:`SelectPhaseSpace`: Adds the resulting event-mask as a new
      variable to the data.

      :meth:`selection.filters.build_selection`: Handles the actual creation of
      the event selection, based on the configuration.

      `numexpr <https://numexpr.readthedocs.io/en/latest/>`_: which is used for
      the internal expression handling.

    """
    def __init__(self, name, out_dir, selection_file=None, keep_unique_id=False,
                 selection=None, counter=True, weights=None):
        self.name = name
        self.out_dir = out_dir
        self.keep_unique_id = keep_unique_id
        if not selection and not selection_file:
            raise BadCutflowConfig("{}: Neither selection nor selection_file specified".format(self.name))
        if selection and selection_file:
            raise BadCutflowConfig("{}: Both selection and selection_file given. Choose one!".format(self.name))

        if selection_file:
            selection = _load_selection_file(self.name, selection_file)

        self._counter = counter
        if not self._counter:
            msg = self.name + ": "
            msg += "Optimisations for when no cut-flow counter is required aren't implemented"
            raise NotImplementedError(msg)
        self._weights = None
        if self._counter:
            self._weights = _create_weights(self.name, weights)

        self.selection = build_selection(self.name, selection, weights=list(self._weights.values()))

    def collector(self):
        outfilename = "cuts_"
        outfilename += self.name + "-"
        outfilename += ".".join(self._weights.keys())
        outfilename += ".csv"
        outfilename = os.path.join(self.out_dir, outfilename)
        return Collector(outfilename, self.keep_unique_id)

    def event(self, chunk):
        is_mc = chunk.config.dataset.eventtype == "mc"
        new_mask = self.selection(chunk.tree, is_mc)
        chunk.tree.apply_mask(new_mask)

    def merge(self, rhs):
        self.selection.merge(rhs.selection)


class SelectPhaseSpace(CutFlow):
    """Creates an event-mask and adds it to the data-space.

    This is identical to the :class:`CutFlow` class, except that the resulting
    mask is added to the list of variables in the data-space, rather than being
    used directly to remove events.  This allows multiple "regions" to be defined
    using different CutFlows in a single configuration.

    Parameters:
      region_name: The name given to the resulting mask when added to back to
        the data-space.

    See Also:
      :class:`CutFlow`: for a description of the other parameters.
    """
    def __init__(self, name, out_dir, region_name, **kwargs):
        super(SelectPhaseSpace, self).__init__(name, out_dir, **kwargs)
        self.region_name = region_name

    def event(self, chunk):
        is_mc = chunk.config.dataset.eventtype == "mc"
        new_mask = self.selection(chunk.tree, is_mc)
        chunk.tree.new_variable(self.region_name, new_mask)



"""
Chop up those trees into nice little tables and dataframes
"""
from __future__ import print_function
import sys
from .help import help_stages
import fast_flow.v1 as fast_flow
import fast_curator
import logging
import atuproot.atuproot_main as atup
from .event_builder import EventBuilder
from atsge.build_parallel import build_parallel
from .utils import mkdir_p
atup.EventBuilder = EventBuilder
atup.build_parallel = build_parallel
logging.getLogger(__name__).setLevel(logging.INFO)


class DummyCollector():
    def collect(self, *args, **kwargs):
        pass


def process_args(args=None):
    from argparse import ArgumentParser, Action

    class StagesHelp(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            full_output = option_string == "--help-stages-full"
            help_stages(values, full_output=full_output)
            sys.exit(0)

    parser = ArgumentParser(description=__doc__)
    parser.add_argument("dataset_cfg", type=str,
                        help="Dataset config to run over")
    parser.add_argument("sequence_cfg", type=str,
                        help="Config for how to process events")
    parser.add_argument("--outdir", default="output", type=str,
                        help="Where to save the results")
    parser.add_argument("--mode", default="multiprocessing", type=str,
                        help="Which mode to run in (multiprocessing, htcondor, sge)")
    parser.add_argument("--ncores", default=0, type=int,
                        help="Number of cores to run on")
    parser.add_argument("--nblocks-per-dataset", default=-1, type=int,
                        help="Number of blocks per dataset")
    parser.add_argument("--nblocks-per-sample", default=-1, type=int,
                        help="Number of blocks per sample")
    parser.add_argument("--blocksize", default=1000000, type=int,
                        help="Number of events per block")
    parser.add_argument("--quiet", default=False, action='store_true',
                        help="Keep progress report quiet")
    parser.add_argument("--profile", default=False, action='store_true',
                        help="Profile the code")
    parser.add_argument("--help-stages", nargs="?", default=None, action=StagesHelp,
                        metavar="stage-name-regex",
                        help="Print help specific to the available stages")
    parser.add_argument("--help-stages-full", action=StagesHelp, metavar="stage",
                        help="Print the full help specific to the available stages")
    return parser.parse_args()


def main(args=None):
    args = process_args(args)

    sequence = fast_flow.read_sequence_yaml(args.sequence_cfg, output_dir=args.outdir)

    datasets = fast_curator.read.from_yaml(args.dataset_cfg)

    mkdir_p(args.outdir)

    process = atup.AtUproot(args.outdir,
                            quiet=args.quiet,
                            parallel_mode=args.mode,
                            process=args.ncores,
                            max_blocks_per_dataset=args.nblocks_per_dataset,
                            max_blocks_per_process=args.nblocks_per_sample,
                            nevents_per_block=args.blocksize,
                            profile=args.profile,
                            profile_out_path="profile.txt",
                            )

    sequence = [(s, s.collector() if hasattr(s, "collector") else DummyCollector()) for s in sequence]
    ret_val = process.run(datasets, sequence)
    print(ret_val)
    return 0


if __name__ == "__main__":
    main()



"""
Summarize the data by producing binned and possibly weighted counts of the data.
"""
import os
import pandas as pd
from . import binning_config as cfg


class Collector():
    def __init__(self, filename, dataset_col, binnings):
        self.filename = filename
        self.dataset_col = dataset_col
        self.binnings = binnings

    def collect(self, dataset_readers_list):
        if len(dataset_readers_list) == 0:
            return

        output = self._prepare_output(dataset_readers_list)
        output.to_csv(self.filename, float_format="%.17g")

    def _prepare_output(self, dataset_readers_list):
        dataset_readers_list = [(d, [r.contents for r in readers])
                                for d, readers in dataset_readers_list if readers]
        if not dataset_readers_list:
            return None

        if self.dataset_col:
            output = _merge_dataframes(dataset_readers_list)
        else:
            output = _add_dataframes(dataset_readers_list)
        if self.binnings:
            output = densify_dataframe(output, self.binnings)
        return output


def _merge_dataframes(dataset_readers_list):
    all_dfs = []
    keys = []
    for dataset, readers in dataset_readers_list:
        dataset_df = readers[0]
        for df in readers[1:]:
            if df is None:
                continue
            dataset_df = dataset_df.add(df, fill_value=0.)
        all_dfs.append(dataset_df)
        keys.append(dataset)
    final_df = pd.concat(all_dfs, keys=keys, names=['dataset'], sort=True)
    return final_df


def _add_dataframes(dataset_readers_list):
    final_df = None
    for _, readers in dataset_readers_list:
        for df in readers:
            if df is None:
                continue
            if final_df is None:
                final_df = df
                continue
            final_df = final_df.add(df, fill_value=0.)
    return final_df


def densify_dataframe(in_df, binnings):
    in_index = in_df.index
    index_values = []
    for dim in in_index.names:
        bins = binnings.get(dim, None)
        if bins is None:
            index_values.append(in_index.unique(dim))
            continue
        index_values.append(pd.IntervalIndex.from_breaks(bins, closed="left"))
    out_index = pd.MultiIndex.from_product(index_values, names=in_index.names)
    out_df = in_df.reindex(index=out_index, copy=False)
    return out_df


class BinnedDataframe():
    """Produces a binned dataframe (a multi-dimensional histogram).

    def __init__(self, name, out_dir, binning, weights=None, dataset_col=False):

    Parameters:
      binning (list[dict]): A list of dictionaries describing the variables to
        bin on, and how they should be binned.  Each of these dictionaries can
        contain the following:

        +-----------+----------------+--------------------------------------------------------------+
        | Parameter | Default        | Description                                                  |
        +===========+================+==============================================================+
        | ``in``    |                | The name of the attribute on the event to use.               |
        +-----------+----------------+--------------------------------------------------------------+
        | ``out``   | same as ``in`` | The name of the column to be filled in the output dataframe. |
        +-----------+----------------+--------------------------------------------------------------+
        | ``bins``  | ``None``       | | Must be either ``None`` or a dictionary.  If a dictionary, |
        |           |                |   it must contain one of the follow sets of                  |
        |           |                | | key-value pairs:                                           |
        |           |                | |   1. ``nbins``, ``low``, ``high``: which are used to       |
        |           |                |     produce a list of bin edges equivalent to:               |
        |           |                | |      ``numpy.linspace(low, high, nbins + 1)``              |
        |           |                | |   2. ``edges``: which is treated as the list of bin        |
        |           |                |     edges directly.                                          |
        |           |                | | If set to ``None``, then the input variable is assumed     |
        |           |                |   to already be categorical (ie. binned or discrete)         |
        +-----------+----------------+--------------------------------------------------------------+

      weights (str or list[str], dict[str, str]): How to weight events in the
        output table.  Must be either a single variable, a list of
        variables, or a dictionary where the values are variables in the data and
        keys are the column names that these weights should be called in the
        output tables.
      dataset_col (bool): adds an extra binning column with the name for each dataset.
      pad_missing (bool): If ``False``, any bins that don't contain data are
        excluded from the stored dataframe.  Leaving this ``False`` can save
        some disk-space and improve processing time, particularly if the bins are
        only very sparsely filled.

    Other Parameters:
      name (str):  The name of this stage (handled automatically by fast-flow)
      out_dir (str):  Where to put the summary table (handled automatically by
          fast-flow)

    Raises:
      BadBinnedDataframeConfig: If there is an issue with the binning description.

    """

    def __init__(self, name, out_dir, binning, weights=None, dataset_col=True, pad_missing=False):
        self.name = name
        self.out_dir = out_dir
        ins, outs, binnings = cfg.create_binning_list(self.name, binning)
        self._bin_dims = ins
        self._out_bin_dims = outs
        self._binnings = binnings
        self._dataset_col = dataset_col
        self._weights = cfg.create_weights(self.name, weights)
        self._pad_missing = pad_missing
        self.contents = None

    def collector(self):
        outfilename = "tbl_"
        if self._dataset_col:
            outfilename += "dataset."
        outfilename += ".".join(self._out_bin_dims)
        outfilename += "--" + self.name
        outfilename += ".csv"
        outfilename = os.path.join(self.out_dir, outfilename)
        binnings = None
        if self._pad_missing:
            binnings = dict(zip(self._out_bin_dims, self._binnings))
        return Collector(outfilename, self._dataset_col, binnings=binnings)

    def event(self, chunk):
        if chunk.config.dataset.eventtype == "mc":
            weights = list(self._weights.values())
            all_inputs = self._bin_dims + weights
        else:
            weights = None
            all_inputs = self._bin_dims

        data = chunk.tree.pandas.df(all_inputs)

        binned_values = _bin_values(data, dimensions=self._bin_dims,
                                    binnings=self._binnings,
                                    weights=weights,
                                    out_weights=self._weights.keys(),
                                    out_dimensions=self._out_bin_dims)
        if self.contents is None:
            self.contents = binned_values
        else:
            self.contents = self.contents.add(binned_values, fill_value=0)

        return True

    def merge(self, rhs):
        if rhs.contents is None or len(rhs.contents) == 0:
            return
        if self.contents is None:
            self.contents = rhs.contents
            return
        self.contents = self.contents.add(rhs.contents, fill_value=0)


_count_label = "n"
_weight_labels = ["sumw", "sumw2"]


def _make_column_labels(weights):
    weight_labels = [w + ":" + l for l in _weight_labels for w in weights]
    return [_count_label] + weight_labels


def _bin_values(data, dimensions, binnings, weights, out_dimensions=None, out_weights=None):
    if not out_dimensions:
        out_dimensions = dimensions
    if not out_weights:
        out_weights = weights

    final_bin_dims = []
    for dimension, binning in zip(dimensions, binnings):
        if binning is None:
            final_bin_dims.append(dimension)
            continue
        out_dimension = dimension + "_bins"
        data[out_dimension] = pd.cut(data[dimension], binning, right=False)
        final_bin_dims.append(out_dimension)

    if weights:
        weight_sq_dims = [w + "_squared" for w in weights]
        data[weight_sq_dims] = data[weights] ** 2

    bins = data.groupby(final_bin_dims)
    counts = bins.size()

    if weights:
        sums = bins[weights].sum()
        sum_sqs = bins[weight_sq_dims].sum()
        histogram = pd.concat([counts, sums, sum_sqs], axis="columns")
        histogram.columns = _make_column_labels(out_weights)
    else:
        histogram = counts.to_frame(_count_label)

    histogram.index.set_names(out_dimensions, inplace=True)
    return histogram



"""
"""
import six
import numpy as np
from awkward import JaggedArray
from ..expressions import get_branches, evaluate
from .reductions import get_pandas_reduction, get_awkward_reduction


class BadVariablesConfig(Exception):
    pass


class Define():
    """Creates new variables using a string-based expression.

    There are two types of expressions:

      * Simple formulae, and
      * Reducing formulae.

    The essential difference, unfortunately, is an internal one: simple
    expressions are nearly directly handled by `numexpr
    <https://numexpr.readthedocs.io/en/latest/>`_, whereas reducing expressions
    add a layer on top.

    From a users perspective, however, simple expressions are those that
    preserve the dimensionality of the input.  If one of the input variables
    represents a list of values for each event  (whose length might vary), then
    the output will contain an equal-length list of values for each event.

    If, however, a reducing expression is used, then there will be one less
    dimension on the resulting variable.  In this case, if an input variable
    has a list of values for each event, the result of the expression will only
    contain a single value per event.

    Parameters:
      variables (list[dictionary]):  A list of single-length dictionaries whose
          key is the name of the resulting variable, and whose value is the
          expression to create it.


    Other Parameters:
      name (str):  The name of this stage (handled automatically by fast-flow)
      out_dir (str):  Where to put the summary table (handled automatically by
          fast-flow)

    Example:
      ::

        variables:
          - Muon_pt: "sqrt(Muon_px**2 + Muon_py**2)"
          - Muon_is_good: (Muon_iso > 0.3) & (Muon_pt > 10)
          - NGoodMuons: {reduce: count_nonzero, formula: Muon_is_good}
          - First_Muon_pt: {reduce: 0, formula: Muon_pt}

    See Also:
      * :mod:`fast_carpenter.define.reductions`-- for how reductions are handled and exactly what is valid.
      * `numexpr <https://numexpr.readthedocs.io/en/latest/>`_: which is used for
        the internal expression handling.

    """

    def __init__(self, name, out_dir, variables):
        self.name = name
        self.out_dir = out_dir
        self._variables = _build_calculations(name, variables, approach="awkward")

    def event(self, chunk):
        for output, expression, reduction, fill_missing in self._variables:
            result = evaluate(chunk.tree, expression)
            if reduction:
                result = reduction(result)

            chunk.tree.new_variable(output, result)
        return True


class DefinePandas():

    def __init__(self, name, out_dir, variables):
        self.name = name
        self.out_dir = out_dir
        self._variables = _build_calculations(name, variables, approach="pandas")

    def event(self, chunk):
        for output, expression, reduction, fill_missing in self._variables:
            branches = get_branches(expression, chunk.tree.allkeys())
            data = chunk.tree.pandas.df(branches)
            result = data.eval(expression)
            if reduction:
                groups = result.groupby(level=0)
                result = reduction(groups)
                array = result.values
            else:
                events = result.index.get_level_values(0).values
                events -= events[0]
                array = JaggedArray.fromparents(events, result.values)

            chunk.tree.new_variable(output, array)
        return True


def _build_calculations(stage_name, variables, approach):
    calculations = []
    for var in variables:
        if not isinstance(var, dict):
            msg = "{}: To define a variable, give me a dictionary for each variable."
            raise RuntimeError(msg.format(stage_name))
        if len(var) != 1:
            msg = "{}: Dictionary to define a variable should have only 1 key-value pair"
            raise RuntimeError(msg.format(stage_name))
        name, config = list(var.items())[0]
        calculations.append(_build_one_calc(stage_name, name, config, approach))
    return calculations


def _build_one_calc(stage_name, name, config, approach):
    reduction = None
    fill_missing = np.nan
    if isinstance(config, six.string_types):
        return name, config, reduction, fill_missing
    if not isinstance(config, dict):
        msg = "{}: To define a new variable need either a string for just a formula or a dictionary"
        raise RuntimeError(msg.format(stage_name))
    if [key for key in config.keys() if key not in ("reduce", "formula", "fill_missing")]:
        msg = "{}: Unknown parameter parsed defining variable '{}'"
        raise RuntimeError(msg.format(stage_name, name))
    if "reduce" in config:
        if approach == "pandas":
            reduction = get_pandas_reduction(stage_name, config["reduce"])
        else:
            reduction = get_awkward_reduction(stage_name, config["reduce"])
    fill_missing = config.get("fill_missing", fill_missing)
    return name, config["formula"], reduction, fill_missing



import numexpr
import tokenize
import awkward
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


__all__ = ["get_branches", "evaluate"]


def get_branches(cut, valid):
    valid = [v.decode("utf-8") for v in valid]

    branches = []
    string = StringIO(cut).readline
    tokens = tokenize.generate_tokens(string)
    current_branch = []
    for toknum, tokval, _, _, _ in tokens:
        if toknum == tokenize.NAME:
            if ".".join(current_branch + [tokval]) in valid:
                current_branch.append(tokval)
                continue
        if tokval == ".":
            continue
        if current_branch:
            branches.append(".".join(current_branch))
            current_branch = []
    return branches


class TreeToDictAdaptor():
    """
    Make an uproot tree look like a dict for numexpr
    """
    def __init__(self, tree):
        self.tree = tree
        self.starts = None
        self.stops = None

    def __getitem__(self, item):
        array = self.tree.array(item)
        starts = getattr(array, "starts", None)
        if starts is not None:
            self.set_starts_stop(starts, array.stops)
            return array.content
        return array

    def __contains__(self, item):
        return item in self.tree

    def __iter__(self):
        for i in self.tree:
            yield i

    def set_starts_stop(self, starts, stops):
        if self.starts is not None:
            if any(self.starts != starts) or any(self.stops != stops):
                raise RuntimeError("Mismatched starts and stops")
        else:
            self.starts = starts
            self.stops = stops


def evaluate(tree, expression):
    adaptor = TreeToDictAdaptor(tree)
    result = numexpr.evaluate(expression, local_dict=adaptor)
    if adaptor.starts is not None:
        result = awkward.JaggedArray(adaptor.starts, adaptor.stops, result)
    return result



# -*- coding: utf-8 -*-

"""Top-level package for fast-carpenter."""

__author__ = """Benjamin Krikler, and F.A.S.T"""
__email__ = 'fast-hep@cern.ch'
__version__ = '0.9.1'


from .define.variables import Define
from .define.systematics import SystematicWeights
from .selection.stage import CutFlow, SelectPhaseSpace
from .summary import BinnedDataframe


__all__ = ["Define", "SystematicWeights", "CutFlow",
           "SelectPhaseSpace", "BinnedDataframe"]


known_stages = [Define, SystematicWeights, CutFlow, SelectPhaseSpace, BinnedDataframe]
